
/*--------------------------------------------------------------------*/
/*--- A header file for all parts of the FlowCheck tool.           ---*/
/*---                                                 fc_include.h ---*/
/*--------------------------------------------------------------------*/

/*
   This file is part of FlowCheck, a heavyweight Valgrind tool for
   detecting leakage of secret information.

   Based on MemCheck, a heavyweight Valgrind tool for detecting memory
   errors.

   Copyright (C) 2000-2007 Julian Seward 
      jseward@acm.org

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.

   The GNU General Public License is contained in the file COPYING.
*/

#ifndef __FC_INCLUDE_H
#define __FC_INCLUDE_H

#define FC_(str)    VGAPPEND(vgFlowCheck_,str)

/*------------------------------------------------------------*/
/*--- Tracking the heap                                    ---*/
/*------------------------------------------------------------*/

/* We want at least a 16B redzone on client heap blocks for Memcheck */
#define FC_MALLOC_REDZONE_SZB    16

/* For malloc()/new/new[] vs. free()/delete/delete[] mismatch checking. */
typedef
   enum {
      FC_AllocMalloc = 0,
      FC_AllocNew    = 1,
      FC_AllocNewVec = 2,
      FC_AllocCustom = 3
   }
   FC_AllocKind;
   
/* Nb: first two fields must match core's VgHashNode. */
typedef
   struct _FC_Chunk {
      struct _FC_Chunk* next;
      Addr         data;            // ptr to actual block
      SizeT        szB : (sizeof(UWord)*8)-2; // size requested; 30 or 62 bits
      FC_AllocKind allockind : 2;   // which wrapper did the allocation
      ExeContext*  where;           // where it was allocated
   }
   FC_Chunk;

/* Memory pool.  Nb: first two fields must match core's VgHashNode. */
typedef
   struct _FC_Mempool {
      struct _FC_Mempool* next;
      Addr          pool;           // pool identifier
      SizeT         rzB;            // pool red-zone size
      Bool          is_zeroed;      // allocations from this pool are zeroed
      VgHashTable   chunks;         // chunks associated with this pool
   }
   FC_Mempool;


extern void* FC_(new_block)  ( ThreadId tid,
                               Addr p, SizeT size, SizeT align, UInt rzB,
                               Bool is_zeroed, FC_AllocKind kind,
                               VgHashTable table);
extern void FC_(handle_free) ( ThreadId tid,
                                Addr p, UInt rzB, FC_AllocKind kind );

extern void FC_(create_mempool)  ( Addr pool, UInt rzB, Bool is_zeroed );
extern void FC_(destroy_mempool) ( Addr pool );
extern void FC_(mempool_alloc)   ( ThreadId tid, Addr pool,
                                   Addr addr, SizeT size );
extern void FC_(mempool_free)    ( Addr pool, Addr addr );
extern void FC_(mempool_trim)    ( Addr pool, Addr addr, SizeT size );
extern void FC_(move_mempool)    ( Addr poolA, Addr poolB );
extern void FC_(mempool_change)  ( Addr pool, Addr addrA, Addr addrB, SizeT size );
extern Bool FC_(mempool_exists)  ( Addr pool );

extern FC_Chunk* FC_(get_freed_list_head)( void );

/* For tracking malloc'd blocks */
extern VgHashTable FC_(malloc_list);

/* For tracking memory pools. */
extern VgHashTable FC_(mempool_list);

/* Shadow memory functions */
extern Bool FC_(check_mem_is_noaccess)( Addr a, SizeT len, Addr* bad_addr );
extern void FC_(make_mem_noaccess) ( Addr a, SizeT len );
extern void FC_(make_mem_undefined)( Addr a, SizeT len );
extern void FC_(make_mem_defined)  ( Addr a, SizeT len );
extern void FC_(make_mem_accessible)(Addr a, SizeT len);
extern void FC_(copy_address_range_state) ( Addr src, Addr dst, SizeT len );

extern void FC_(print_malloc_stats) ( void );

extern void* FC_(malloc)               ( ThreadId tid, SizeT n );
extern void* FC_(__builtin_new)        ( ThreadId tid, SizeT n );
extern void* FC_(__builtin_vec_new)    ( ThreadId tid, SizeT n );
extern void* FC_(memalign)             ( ThreadId tid, SizeT align, SizeT n );
extern void* FC_(calloc)               ( ThreadId tid, SizeT nmemb, SizeT size1 );
extern void  FC_(free)                 ( ThreadId tid, void* p );
extern void  FC_(__builtin_delete)     ( ThreadId tid, void* p );
extern void  FC_(__builtin_vec_delete) ( ThreadId tid, void* p );
extern void* FC_(realloc)              ( ThreadId tid, void* p, SizeT new_size );


/*------------------------------------------------------------*/
/*--- Profiling of memory events                           ---*/
/*------------------------------------------------------------*/

/* Define to collect detailed performance info. */
/* #define FC_PROFILE_MEMORY */

#ifdef FC_PROFILE_MEMORY
#  define N_PROF_EVENTS 500

extern UInt   FC_(event_ctr)[N_PROF_EVENTS];
extern HChar* FC_(event_ctr_name)[N_PROF_EVENTS];

#  define PROF_EVENT(ev, name)                                \
   do { tl_assert((ev) >= 0 && (ev) < N_PROF_EVENTS);         \
        /* crude and inaccurate check to ensure the same */   \
        /* event isn't being used with > 1 name */            \
        if (FC_(event_ctr_name)[ev])                         \
           tl_assert(name == FC_(event_ctr_name)[ev]);       \
        FC_(event_ctr)[ev]++;                                \
        FC_(event_ctr_name)[ev] = (name);                    \
   } while (False);

#else

#  define PROF_EVENT(ev, name) /* */

#endif   /* FC_PROFILE_MEMORY */


/*------------------------------------------------------------*/
/*--- V and A bits (Victoria & Albert ?)                   ---*/
/*------------------------------------------------------------*/

/* The number of entries in the primary map can be altered.  However
   we hardwire the assumption that each secondary map covers precisely
   64k of address space. */
#define SM_SIZE 65536            /* DO NOT CHANGE */
#define SM_MASK (SM_SIZE-1)      /* DO NOT CHANGE */

#define V_BIT_DEFINED         0
#define V_BIT_UNDEFINED       1

#define V_BITS8_DEFINED       0
#define V_BITS8_UNDEFINED     0xFF

#define V_BITS16_DEFINED      0
#define V_BITS16_UNDEFINED    0xFFFF

#define V_BITS32_DEFINED      0
#define V_BITS32_UNDEFINED    0xFFFFFFFF

#define V_BITS64_DEFINED      0ULL
#define V_BITS64_UNDEFINED    0xFFFFFFFFFFFFFFFFULL


/*------------------------------------------------------------*/
/*--- Leak checking                                        ---*/
/*------------------------------------------------------------*/

/* A block is either 
   -- Proper-ly reached; a pointer to its start has been found
   -- Interior-ly reached; only an interior pointer to it has been found
   -- Unreached; so far, no pointers to any part of it have been found. 
   -- IndirectLeak; leaked, but referred to by another leaked block
*/
typedef 
   enum { 
      Unreached    =0, 
      IndirectLeak =1,
      Interior     =2, 
      Proper       =3
  }
  Reachedness;

/* For VALGRIND_COUNT_LEAKS client request */
extern SizeT FC_(bytes_leaked);
extern SizeT FC_(bytes_indirect);
extern SizeT FC_(bytes_dubious);
extern SizeT FC_(bytes_reachable);
extern SizeT FC_(bytes_suppressed);

typedef
   enum {
      LC_Off,
      LC_Summary,
      LC_Full,
   }
   LeakCheckMode;

/* A block record, used for generating err msgs. */
typedef
   struct _LossRecord {
      struct _LossRecord* next;
      /* Where these lost blocks were allocated. */
      ExeContext*  allocated_at;
      /* Their reachability. */
      Reachedness  loss_mode;
      /* Number of blocks and total # bytes involved. */
      SizeT        total_bytes;
      SizeT        indirect_bytes;
      UInt         num_blocks;
   }
   LossRecord;

extern void FC_(do_detect_memory_leaks) (
          ThreadId tid, LeakCheckMode mode,
          Bool (*is_within_valid_secondary) ( Addr ),
          Bool (*is_valid_aligned_word)     ( Addr )
       );

extern void FC_(pp_LeakError)(UInt n_this_record, UInt n_total_records,
                              LossRecord* l);
                          

/*------------------------------------------------------------*/
/*--- Errors and suppressions                              ---*/
/*------------------------------------------------------------*/

extern void FC_(record_free_error)            ( ThreadId tid, Addr a ); 
extern void FC_(record_illegal_mempool_error) ( ThreadId tid, Addr a );
extern void FC_(record_freemismatch_error)    ( ThreadId tid, FC_Chunk* mc );
extern Bool FC_(record_leak_error)            ( ThreadId tid,
                                                UInt n_this_record,
                                                UInt n_total_records,
                                                LossRecord* lossRecord,
                                                Bool print_record );

/*------------------------------------------------------------*/
/*--- Command line options + defaults                      ---*/
/*------------------------------------------------------------*/

/* Allow loads from partially-valid addresses?  default: YES */
extern Bool FC_(clo_partial_loads_ok);

/* Max volume of the freed blocks queue. */
extern Long FC_(clo_freelist_vol);

/* Do leak check at exit?  default: NO */
extern LeakCheckMode FC_(clo_leak_check);

/* How closely should we compare ExeContexts in leak records? default: 2 */
extern VgRes FC_(clo_leak_resolution);

/* In leak check, show reachable-but-not-freed blocks?  default: NO */
extern Bool FC_(clo_show_reachable);

/* Assume accesses immediately below %esp are due to gcc-2.96 bugs.
 * default: NO */
extern Bool FC_(clo_workaround_gcc296_bugs);

/* Do undefined value checking? "No" gives Addrcheck-style behaviour, ie.
 * faster but fewer errors found.  Note that although Addrcheck had 1 bit
 * per byte overhead vs the old Memcheck's 9 bits per byte, with this mode
 * and compressed V bits, no memory is saved with this mode -- it's still
 * 2 bits per byte overhead.  This is a little wasteful -- it could be done
 * with 1 bit per byte -- but lets us reuse the many shadow memory access
 * functions.  Note also that in this mode the secondary V bit table is
 * never used.
 *
 * default: YES */
extern Bool FC_(clo_undef_value_errors);

/* Fill malloc-d/free-d client blocks with a specific value?  -1 if
   not, else 0x00 .. 0xFF indicating the fill value to use.  Can be
   useful for causing programs with bad heap corruption to fail in
   more repeatable ways.  Note that malloc-filled and free-filled
   areas are still undefined and noaccess respectively.  This merely
   causes them to contain the specified values. */
extern Int FC_(clo_malloc_fill);
extern Int FC_(clo_free_fill);

/* Treat data read from non-world-readable files as tainted?
 * default: NO */
extern Bool FC_(clo_private_files_are_secret);

/* Treat data that's unpredictable, like /dev/urandom, the PID, and
 * the time of day, as `secret' (for use in testing RNGs, etc.)?
 * default: NO */
extern Bool FC_(clo_unpredictable_is_secret);

/* Treat data read from the standard input as tainted?
 * default: NO */
extern Bool FC_(clo_stdin_is_secret);

/* Report leaks with a backtrace and other Memcheck-like details?
 * default: NO */
extern Bool  FC_(clo_detailed_leak_report);

/* Abort execution after this many bits have been leaked
 * default: effectively infinite */
extern Long  FC_(clo_max_bits_leaked);

/* Output a graph showing all the operations on secrets along with
 * edge capacities?
 * default: NO */
extern Bool  FC_(clo_trace_secret_graph);

/* Mark all bits generated inside an enclosure region as secret?
 * default: NO */
extern Bool  FC_(clo_taint_all_enclosed);

/* Print a warning message every time an enclosure region reverts a *
 * memory location on exit?
 * default: NO */
extern Bool  FC_(clo_trace_reverts);

/* Taint the modified bits every time an enclosure region reverts a *
 * memory location on exit?
 * default: NO */
extern Bool  FC_(clo_taint_reverts);

extern Int  FC_(clo_revert_notbelow);

/* To what extent should the flow graph be collapsed by combining
   similar edges? Currently defined values:

   0   - do not collapse at all (edges will be output incrementally)
   10  - collapse edges with the same source location and call stack
         (fully context sensitive)
   50  - collapse edges with the same source location and caller
         (one level context sensitive)
   100 - collapse edges with the same source location, no matter
         what call stack (context insensitive)
   default: 10 */
extern Int FC_(clo_folding_level);

/* Run the maximum flow computation after each output? */
extern Bool FC_(clo_incremental);

/* File to store the produced graph in */
extern Char *FC_(clo_graph_file);

/* Maximum flow program to pass it to */
extern Char *FC_(clo_max_flow_program);

/*------------------------------------------------------------*/
/*--- Variables                                            ---*/
/*------------------------------------------------------------*/

extern Long FC_(total_bits_leaked);
extern int  FC_(enclosure_mode);

/*------------------------------------------------------------*/
/*--- Rollback                                             ---*/
/*------------------------------------------------------------*/

extern int  FC_(rollback_mode);

extern Bool FC_(is_escaping)(Addr addr);

/*------------------------------------------------------------*/
/*--- Instrumentation                                      ---*/
/*------------------------------------------------------------*/

/* Functions defined in fc_main.c */
extern VG_REGPARM(3) void FC_(helperc_complain_undef) ( HWord, UInt,
							UInt, ULong, ULong );
extern void FC_(helperc_value_check8_fail) ( ULong, ULong );
extern VG_REGPARM(2) void FC_(helperc_value_check4_fail) ( UInt, UInt, ULong,
							   ULong );
extern VG_REGPARM(2) void FC_(helperc_value_check1_fail) ( UChar, UChar,
							   ULong, ULong);
extern void FC_(helperc_value_check0_fail) ( ULong, ULong );

extern VG_REGPARM(1) void FC_(helperc_STOREV64be) ( Addr, ULong );
extern VG_REGPARM(1) void FC_(helperc_STOREV64le) ( Addr, ULong );
extern VG_REGPARM(2) void FC_(helperc_STOREV32be) ( Addr, UWord );
extern VG_REGPARM(2) void FC_(helperc_STOREV32le) ( Addr, UWord );
extern VG_REGPARM(2) void FC_(helperc_STOREV16be) ( Addr, UWord );
extern VG_REGPARM(2) void FC_(helperc_STOREV16le) ( Addr, UWord );
extern VG_REGPARM(2) void FC_(helperc_STOREV8)   ( Addr, UWord );

extern VG_REGPARM(1) ULong FC_(helperc_LOADV64be) ( Addr );
extern VG_REGPARM(1) ULong FC_(helperc_LOADV64le) ( Addr );
extern VG_REGPARM(1) UWord FC_(helperc_LOADV32be) ( Addr );
extern VG_REGPARM(1) UWord FC_(helperc_LOADV32le) ( Addr );
extern VG_REGPARM(1) UWord FC_(helperc_LOADV16be) ( Addr );
extern VG_REGPARM(1) UWord FC_(helperc_LOADV16le) ( Addr );
extern VG_REGPARM(1) UWord FC_(helperc_LOADV8)    ( Addr );

extern VG_REGPARM(1) void FC_(save_for_rollback8) (Addr location);
extern VG_REGPARM(1) void FC_(save_for_rollback16)(Addr location);
extern VG_REGPARM(1) void FC_(save_for_rollback32)(Addr location);
extern VG_REGPARM(1) void FC_(save_for_rollback64)(Addr location);

extern void FC_(helperc_MAKE_STACK_UNINIT) ( Addr base, UWord len );

/* Functions defined in fc_translate.c */
extern
IRSB* FC_(instrument) ( VgCallbackClosure* closure,
                        IRSB* bb_in, 
                        VexGuestLayout* layout, 
                        VexGuestExtents* vge,
                        IRType gWordTy, IRType hWordTy );

extern
IRSB* FC_(final_tidy) ( IRSB* );

#endif /* ndef __FC_INCLUDE_H */

/*--------------------------------------------------------------------*/
/*--- end                                                          ---*/
/*--------------------------------------------------------------------*/

