/*--------------------------------------------------------------------*/
/*--- Runtime operations related to constructing a data-flow       ---*/
/*--- graph of program execution.                                  ---*/
/*---                                              trace_runtime.c ---*/
/*--------------------------------------------------------------------*/

/*
   This file is part of FlowCheck, a heavyweight Valgrind tool for
   detecting leakage of secret information.

   By Stephen McCamant, MIT CSAIL Program Analsis group, Copyright (C)
   2007-2008.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.

   The GNU General Public License is contained in the file COPYING.
*/

#include "pub_tool_basics.h"
#include "pub_tool_aspacemgr.h"
#include "pub_tool_hashtable.h"     // For fc_include.h
#include "pub_tool_debuginfo.h"
#include "pub_tool_machine.h"
#include "pub_tool_mallocfree.h"
#include "pub_tool_threadstate.h"
#include "pub_tool_tooliface.h"
#include "pub_tool_oset.h"
#include "pub_tool_vki.h"
#include "pub_tool_libcbase.h"
#include "pub_tool_libcfile.h"
#include "pub_tool_libcprint.h"
#include "pub_tool_libcassert.h"
#include "fc_include.h"
#include "fc_translate.h" /* To get FC_(tags_guest_offset) */
#include "trace_runtime.h"
#include "fold.h"

#define MORE_DEBUG

#if VG_WORDSIZE == 4
#define HIGH_BITS(addr) ((addr) >> 16)
#define LOW_BITS(addr) ((addr) & 0xffff)
#define PAGE_SIZE 65536
#define TABLE_SIZE 65536
#else 
#error Only 32-bit architectures supported
#endif

static ULong zero_page[PAGE_SIZE];

static ULong *page_table[TABLE_SIZE];

#define MAX_FLOW_EXCEPTIONS 30
struct flow_region {
   Addr start;
   SizeT size;
   ULong src_tag;
   int num_exceptions;
   Addr exceptions[MAX_FLOW_EXCEPTIONS];
};

#define MAX_FLOW_REGIONS 40
static struct flow_region flow_regions[MAX_FLOW_REGIONS];
static int num_flow_regions;

#define MIN(a, b) ((a) < (b) ? (a) : (b))
#define MAX(a, b) ((a) > (b) ? (a) : (b))

/* Hash of the calling context, a la "Probabilistic Calling Context",
   Bond and McKinley OOPSLA 2007 */
static ULong callstack_hash = 0;
static Addr last_callsite;

/* Tourist information: graph size before collapsing */
static ULong total_nodes = 0;
static ULong total_edges = 0;

/* File descriptor to which graph edges should be output. The value of 2
   is special, since that's where Valgrind's printf goes. */
static Int graph_output_fd;

void output_graph_stats(void) {
   VG_(printf)("Uncollapsed graph had %lld nodes, %lld edges\n",
	       total_nodes, total_edges);
   if (graph_output_fd != 2) {
      VG_(close)(graph_output_fd);
   }
}

static Word edge_compare(void *key, void *elem) {
   ULong *key_llp = key;
   struct edge_info *edge = (struct edge_info *)elem;
   if (key_llp[0] < edge->location)
      return -1;
   else if (key_llp[0] > edge->location)
      return 1;
   else if (key_llp[1] < edge->context)
      return -1;
   else if (key_llp[1] > edge->context)
      return 1;
   else {
      tl_assert(key_llp[0] == edge->location);
      tl_assert(key_llp[1] == edge->context);
      return 0;
   }
   /*return (Word)(key_llp[0] - edge->location)
     || (Word)(key_llp[1] - edge->context);*/
}

void init_tag_memory(void) {
   int i;
   for (i = 0; i < TABLE_SIZE; i++)
      page_table[i] = zero_page;
   edges = VG_(OSetGen_Create)(offsetof(struct edge_info, location),
			       edge_compare, VG_(malloc), VG_(free));
   if (FC_(clo_graph_file)) {
      SysRes sr;
      sr = VG_(open)(FC_(clo_graph_file), VKI_O_WRONLY|VKI_O_CREAT|VKI_O_TRUNC,
		     0666);
      tl_assert(!sr.isError);
      graph_output_fd = sr.res;
   } else {
      graph_output_fd = 2;
   }
   if (FC_(clo_folding_level)) {
      fold_init();
      fold_new_tag(SOURCE_TAG);
      fold_new_tag(SINK_TAG);
   }
}

static ULong get_single_tag(Addr addr) {
   /*VG_(printf)("tag[%08x] is %lld (flat)\n", addr,
     page_table[HIGH_BITS(addr)][LOW_BITS(addr)]);*/
   return page_table[HIGH_BITS(addr)][LOW_BITS(addr)];
}

static void set_single_tag(Addr addr, ULong tag) {
   if (page_table[HIGH_BITS(addr)] == zero_page) {
      int i;
      ULong *page = VG_(am_shadow_alloc)(PAGE_SIZE * sizeof(ULong));
      if (!page)
	 VG_(out_of_memory_NORETURN)("flowcheck: allocate new tag page", 
				     PAGE_SIZE * sizeof(ULong));
      for (i = 0; i < PAGE_SIZE; i++)
	 page[i] = 0;
      page_table[HIGH_BITS(addr)] = page;
   }
   /*VG_(printf)("tag[%08x] := %lld (flat)\n", addr, tag);*/
   page_table[HIGH_BITS(addr)][LOW_BITS(addr)] = tag;
}

void mark_special_tags(void);

void mark_memory_tags(void) {
   int i;
   for (i = 0; i < TABLE_SIZE; i++) {
      int j;
      ULong *page = page_table[i];
      if (page == zero_page)
	 continue;
      for (j = 0; j < PAGE_SIZE; j++)
	 if (page[j] != 0)
	    mark_tag_in_use(page[j]);
   }
   for (i = 0; i < num_flow_regions; i++) {
      mark_tag_in_use(flow_regions[i].src_tag);
   }
   mark_special_tags();
}

static void new_edge(ULong src, ULong dest, int capacity,
		     const char *type, ULong location) {
   tl_assert(src != SINK_TAG && dest != SOURCE_TAG);
   /* Graph should be a DAG, but node numbers aren't always increasing */
   /* tl_assert(dest == SINK_TAG || dest > src); */
   total_edges++;
   if (capacity == 0) /* can discard immediately */
      return;
   if (FC_(clo_folding_level) == 0) {
      Addr eip = VG_(get_IP)(VG_(get_running_tid)());
      Char buf[256];
      VG_(describe_IP)(eip, buf, sizeof(buf));
      if (src == SOURCE_TAG)
	 src = 0;
      if (dest == SINK_TAG)
	 dest = -1;
      if (graph_output_fd == 2) {
	 VG_(printf)("%lld %lld %d %llx:%llx %s (%s)\n", src, dest,
		     (Int)capacity, location, callstack_hash, buf, type);
      } else {
	 fdprintf(graph_output_fd, "%lld %lld %d %llx:%llx %s (%s)\n",
		  src, dest, (Int)capacity, location,
		  callstack_hash, buf, type);
      }
   } else {
      ULong key[2];
      struct edge_info *edge;
      ULong context;
      ULong loc;
      if (FC_(clo_folding_level) <= 10) {
	 context = callstack_hash;
      } else if (FC_(clo_folding_level) <= 50) {
	 context = last_callsite;
      } else {
	 context = 0;
      }
      /*loc = (location << 8) | ((int)type & 0xff);*/
      loc = location;
      key[0] = loc;
      key[1] = context;
      edge = VG_(OSetGen_Lookup)(edges, &key);
      if (!edge) {
	 Addr eip = VG_(get_IP)(VG_(get_running_tid)());
	 Char buf[256];
	 edge = VG_(OSetGen_AllocNode)(edges, sizeof(struct edge_info));
	 edge->location = loc;
	 edge->context = context;
	 edge->eip = eip;
	 VG_(describe_IP)(eip, buf, sizeof(buf));
	 edge->eip_descr = VG_(strdup)(buf);
	 edge->type_descr = type;
	 edge->src_tag = edge->dest_tag = NO_TAG;
	 edge->total_capacity = 0;
	 VG_(OSetGen_Insert)(edges, edge);
      }
      fold_new_edge(src, dest, capacity, edge);
   }
}

/* Used only when FC_(clo_folding_level) == 0. In that case, we assign
   new tags freely and never garbage collect them. But when we're folding,
   we need to keep state for each tag; then we keep only a limited number
   around at once, and reuse indices by garbage collecting. */
static ULong next_tag = FIRST_TAG;

static ULong get_next_tag(void) {
   total_nodes++;
   if (FC_(clo_folding_level))
      return get_free_tag();
   else
      return next_tag++;
}

static void add_flow_region(Addr start, SizeT size, ULong src_tag,
			    ULong location) {
   int i;
   Addr end = start + size;
   for (i = 0; i < num_flow_regions; i++) {
      Addr old_start = flow_regions[i].start;
      int old_size =  flow_regions[i].size;
      Addr old_end = old_start + old_size;
      if (!(end < old_start) && !(old_end < start)) {
         /* overlapping or adjacent regions: merge */
         Addr new_end;
         int j;
	 ULong new_tag = get_next_tag();
         flow_regions[i].start = MIN(start, old_start);
         new_end = MAX(end, old_end);
         flow_regions[i].size = new_end - flow_regions[i].start;
         /*if (flow_regions[i].start != old_start || new_end != old_end)
            VG_(printf)("Expanding flow region %d to 0x%08x to 0x%08x"
                        " (size %d)\n",
                        i, flow_regions[i].start, end, flow_regions[i].size);
	 else
            VG_(printf)("Renewing flow region %d to 0x%08x to 0x%08x"
                        " (size %d)\n",
                        i, flow_regions[i].start, end, flow_regions[i].size);*/
         for (j = 0; j < flow_regions[i].num_exceptions; j++) {
            Addr a = flow_regions[i].exceptions[j];
            if (a >= start && a < end) {
               Addr old_num = flow_regions[i].num_exceptions;
               /*VG_(printf)("Removing exception for 0x%08x\n", a);*/
               flow_regions[i].exceptions[j] = 
                  flow_regions[i].exceptions[old_num - 1];
               --flow_regions[i].num_exceptions;
	       j--;
            }
         }
	 new_edge(flow_regions[i].src_tag, new_tag, old_size * 8,
		     "flow region cascade old", location);
	 new_edge(src_tag, new_tag, size * 8, "flow region cascade new",
		     location + 1);
	 flow_regions[i].src_tag = new_tag;
         return;
      }
   }
   i = num_flow_regions++;
   tl_assert(num_flow_regions <= MAX_FLOW_REGIONS);
   flow_regions[i].start = start;
   flow_regions[i].size = size;
   flow_regions[i].num_exceptions = 0;
   flow_regions[i].src_tag = src_tag;
   /*VG_(printf)("Adding flow region %d from 0x%08x to 0x%08x (size %d)\n",
     i, start, end, size);*/
}

static void partial_give_up_on_flow_region(int i, SizeT size, ULong location) {
   Addr a;
   Addr end = flow_regions[i].start + size;
   Bool partial = (size < flow_regions[i].size);
   /*const char *partial_str = partial ? "partial " : "";*/
   int j;

   /*if (partial) {
      VG_(printf)("Scaling back flow region %d\n", i);      
   } else {
      VG_(printf)("Destroying flow region %d\n", i);
   }
   VG_(printf)("Exceptions were:\n");
   for (j = 0; j < flow_regions[i].num_exceptions; j++) {
      Addr addr = flow_regions[i].exceptions[j];
      VG_(printf)("    0x%08x\n", addr);
      }*/

   a = flow_regions[i].start;
   /*VG_(printf)("On %sdestruction, tainting from 0x%08x to 0x%08x\n",
     partial_str, a, end);*/
   for (; a < end; a++) {
      Bool is_exception = False;
      for (j = 0; j < flow_regions[i].num_exceptions; j++) {
	 Addr a2 = flow_regions[i].exceptions[j];
	 tl_assert(a2 >= flow_regions[i].start && a2 < end);
	 if (a == a2) {
	    /* VG_(printf)("  except 0x%08x\n", a);*/
	    is_exception = True;
	    break;
	 }
      }
      if (!is_exception) {
	 ULong old_tag = get_single_tag(a);
	 ULong new_tag = get_next_tag();
	 if (old_tag)
	    new_edge(old_tag, new_tag, 8, partial ? 
			"old value in partial give up" :
			"old value in give up", location + 2 * partial);
	 new_edge(flow_regions[i].src_tag, new_tag, 8, partial ?
		     "enclosed output in partial give up" :
		     "enclosed output in give up", location + 2*partial + 1);
	 set_single_tag(a, new_tag);
      }
   }
   flow_regions[i].num_exceptions = 0;
   flow_regions[i].start += size;
   flow_regions[i].size -= size;
}

static void give_up_on_flow_region(int i, ULong location) {
  partial_give_up_on_flow_region(i, flow_regions[i].size, location);
   flow_regions[i] = flow_regions[--num_flow_regions];
}

static void add_flow_exception(int i, Addr addr, ULong location) {
   if (addr == flow_regions[i].start) {
      flow_regions[i].start++;
      flow_regions[i].size--;
      return;
   }
   flow_regions[i].exceptions[flow_regions[i].num_exceptions++] = addr;
   if (flow_regions[i].num_exceptions == MAX_FLOW_EXCEPTIONS) {
      int j;
      Addr max_exception = flow_regions[i].exceptions[0];
      SizeT partial_size;
      for (j = 1; j < flow_regions[i].num_exceptions; j++) {
	 if (flow_regions[i].exceptions[j] > max_exception)
	    max_exception = flow_regions[i].exceptions[j];
      }
      /* VG_(printf)("Largest exeception is 0x%08x\n", max_exception);*/
      partial_size = max_exception - flow_regions[i].start + 1;
      if (2 * partial_size < flow_regions[i].size)
	 partial_give_up_on_flow_region(i, partial_size, location);
      else 
	 give_up_on_flow_region(i, location);
   }
}

ULong get_tag(Addr addr, ULong location) {
   int i;
   for (i = 0; i < num_flow_regions; i++) {
      /*VG_(printf)("Checking 0x%08x versus 0x%08x and 0x%08x (flow)\n",
		  addr, flow_regions[i].start,
		  flow_regions[i].start + flow_regions[i].size);*/
      if (addr >= flow_regions[i].start &&
	  addr < flow_regions[i].start + flow_regions[i].size) {
	 int j;
	 ULong old_tag, new_tag;
	 for (j = 0; j < flow_regions[i].num_exceptions; j++)
	    if (flow_regions[i].exceptions[j] == addr) {
	       /*VG_(printf)("0x%08x is an exception\n", addr);*/
	       return get_single_tag(addr);
	    }
	 old_tag = get_single_tag(addr);
	 new_tag = get_next_tag();
	 if (old_tag)
	    new_edge(old_tag, new_tag, 8,
			"old value of enclosed output region", location+31);
	 new_edge(flow_regions[i].src_tag, new_tag, 8,
		     "enclosed output in region", location+32);
	 set_single_tag(addr, new_tag);
	 if (flow_regions[i].num_exceptions < MAX_FLOW_EXCEPTIONS / 2)
	    add_flow_exception(i, addr, location);
	 return new_tag;
      }
   }
   return get_single_tag(addr);
}

void set_tag(Addr addr, ULong tag, ULong location) {
   int i;
   for (i = 0; i < num_flow_regions; i++) {
      if (addr >= flow_regions[i].start &&
	  addr < flow_regions[i].start + flow_regions[i].size) {
	 int j;
	 for (j = 0; j < flow_regions[i].num_exceptions; j++)
	    if (flow_regions[i].exceptions[j] == addr) {
	       /*VG_(printf)("0x%08x is already an exception\n", addr);*/
	       return set_single_tag(addr, tag);
	    }
	 add_flow_exception(i, addr, location);
	 return set_single_tag(addr, tag);
      }
   }
   return set_single_tag(addr, tag);
}

void make_freshly_tagged_input(Addr start, SizeT len, ULong location) {
   SizeT i;
   check_enough_tags(len);
   for (i = 0; i < len; i++) {
      ULong tag = get_next_tag();
      set_tag(start + i, tag, location);
      new_edge(SOURCE_TAG, tag, 8, "input byte", location);
   }
}

void make_freshly_tagged_register(ThreadId tid, OffT offset, int capacity,
				  ULong location) {
   ULong tag;
   OffT tag_offset;
   check_enough_tags(1);
   tag = get_next_tag();
   tag_offset = FC_(tags_guest_offset) + 8 * offset;
   VG_(set_shadow_regs_area)(tid, tag_offset, sizeof(ULong), (UChar*)&tag);
   new_edge(SOURCE_TAG, tag, capacity, "input register", location);
}

static ULong leak_cascade = 0;

void count_leaked_tags(Addr start, SizeT len, ULong location) {
   SizeT i;
   ULong next_cascade = 0;
   check_enough_tags(2 * len);
   for (i = 0; i < len; i++) {
      ULong tag = get_tag(start + i, location);
      if (tag) {
	 if (leak_cascade) {
	    ULong join_node = get_next_tag();
	    if (!next_cascade) {
	       next_cascade = get_next_tag();
	       new_edge(leak_cascade, next_cascade, FC_(total_bits_leaked),
			   "global leak cascade", location);
	    }
	    new_edge(tag, join_node, 8, "output byte to join", location+1);
	    new_edge(leak_cascade, join_node, 8, "global leak to output",
			location+2);
	    tag = join_node;
	 }
	 new_edge(tag, SINK_TAG, 8, "output byte", location+3);
      } else if (leak_cascade) {
	new_edge(leak_cascade, SINK_TAG, 8, "global leak on public output",
		    location+4);
      }
   }
   if (next_cascade) {
      leak_cascade = next_cascade;
   }
   if (FC_(clo_incremental))
      run_incremental_max_flow();
}

static ULong combine2(ULong tag1, ULong tag2, ULong location) {
   if (tag1 == tag2) {
      return tag1;
   } else if (!tag1) {
      return tag2;
   } else if (!tag2) {
      return tag1;
   } else {
      ULong new_tag = get_next_tag();
      new_edge(tag1, new_tag, 8, "combine2", location + 1);
      new_edge(tag2, new_tag, 8, "combine2", location + 2);
      return new_tag;
   }
}

static ULong merge2(ULong tag1, ULong tag2, ULong location) {
   if (tag1 == tag2)
      return tag1;
   else
      return combine2(tag1, tag2, location);
}

static ULong combine4(ULong tag1, ULong tag2, ULong tag3, ULong tag4,
		      ULong location) {
   if (tag1 == tag2 && tag2 == tag3 && tag3 == tag4) {
      return tag1;
   } else if (!tag1 && !tag2 && !tag3) {
      return tag4;
   } else if (!tag2 && !tag3 && !tag4) {
      return tag1;
   } else if (!tag1 && !tag3 && !tag4) {
      return tag2;
   } else if (!tag1 && !tag2 && !tag4) {
      return tag3;
   } else {
      ULong new_tag = get_next_tag();
      if (tag1)
	 new_edge(tag1, new_tag, 8, "combine4", location);
      if (tag2)
	 new_edge(tag2, new_tag, 8, "combine4", location);
      if (tag3)
	 new_edge(tag3, new_tag, 8, "combine4", location);
      if (tag4)
	 new_edge(tag4, new_tag, 8, "combine4", location);
      return new_tag;
   }
}

static ULong merge4(ULong tag1, ULong tag2, ULong tag3, ULong tag4,
		    ULong location) {
   if (tag1 == tag2 && tag2 == tag3 && tag3 == tag4)
      return tag1;
   else
      return combine4(tag1, tag2, tag3, tag4, location);
}

ULong FC_(helperc_RESULT_TAG)(ULong tag, UInt width, ULong location) {
   if (width && !tag) {
      VG_(printf)("Missing tag for result at %llx\n", location);
      tl_assert(!width || tag);
   }
   if (!tag) {
      return tag;
   } else {
      ULong new_tag;
      check_enough_tags(1);
      new_tag = get_next_tag();
      new_edge(tag, new_tag, width, "result", location + 10);
      return new_tag;
   }
}

ULong FC_(helperc_UNARY_TAG)(ULong tag, UInt width, ULong location) {
   if (!tag) {
      return tag;
   } else {
      ULong new_tag;
      check_enough_tags(1);
      new_tag = get_next_tag();
      new_edge(tag, new_tag, width, "unary", location + 20);
      return new_tag;
   }
}

ULong FC_(helperc_BINARY_TAGS)(ULong tag1, ULong tag2,
			       UInt width1, UInt width2,
			       ULong location) {
   ULong new_tag;
   tl_assert(!width1 || tag1);
   if (width2 && !tag2) {
      VG_(printf)("Missing tag for second arg at %llx\n", location);
      tl_assert(!width2 || tag2);
   }
   if (!tag1 && !tag2)
      return 0;
   check_enough_tags(1);
   new_tag = get_next_tag();
   if (tag1)
      new_edge(tag1, new_tag, width1, "binary arg 1", location + 1);
   if (tag2)
      new_edge(tag2, new_tag, width2, "binary arg 2", location + 2);
   return new_tag;
}

ULong FC_(helperc_TERNARY_TAGS)(ULong tag1, ULong tag2, ULong tag3,
				UInt width1, UInt width2, UInt width3,
				ULong location) {
   ULong new_tag;
   if (!tag1 && !tag2 && !tag3)
      return 0;
   check_enough_tags(1);
   new_tag = get_next_tag();
   if (tag1)
      new_edge(tag1, new_tag, width1, "ternary arg 1", location + 1);
   if (tag2)
      new_edge(tag2, new_tag, width2, "ternary arg 2", location + 2);
   if (tag3)
      new_edge(tag3, new_tag, width3, "ternary arg 3", location + 3);
   return new_tag;
}

ULong FC_(helperc_FOURARY_TAGS)(ULong tag1, ULong tag2, ULong tag3, ULong tag4,
				UInt width1, UInt width2, UInt width3,
				UInt width4, ULong location) {
   ULong new_tag;
   if (!tag1 && !tag2 && !tag3 && !tag4)
      return 0;
   check_enough_tags(1);
   new_tag = get_next_tag();
   if (tag1)
      new_edge(tag1, new_tag, width1, "fourary arg 1", location + 1);
   if (tag2)
      new_edge(tag2, new_tag, width2, "fourary arg 2", location + 2);
   if (tag3)
      new_edge(tag3, new_tag, width3, "fourary arg 3", location + 3);
   if (tag4)
      new_edge(tag4, new_tag, width4, "fourary arg 4", location + 4);
   return new_tag;
}

ULong FC_(helperc_FIVEARY_TAGS)(ULong tag1, ULong tag2, ULong tag3, ULong tag4,
				ULong tag5,
				UInt width1, UInt width2, UInt width3,
				UInt width4, UInt width5, ULong location) {
   ULong new_tag;
   if (!tag1 && !tag2 && !tag3 && !tag4 && !tag5)
      return 0;
   check_enough_tags(1);
   new_tag = get_next_tag();
   if (tag1)
      new_edge(tag1, new_tag, width1, "fiveary arg 1", location + 1);
   if (tag2)
      new_edge(tag2, new_tag, width2, "fiveary arg 2", location + 2);
   if (tag3)
      new_edge(tag3, new_tag, width3, "fiveary arg 3", location + 3);
   if (tag4)
      new_edge(tag4, new_tag, width4, "fiveary arg 4", location + 4);
   if (tag5)
      new_edge(tag5, new_tag, width5, "fiveary arg 5", location + 5);
   return new_tag;
}

ULong FC_(helper_LOAD_TAG_1) (Addr addr, ULong addrTag, ULong location) {
   ULong tag = get_tag(addr, location);
#ifdef EVEN_MORE_DEBUG
   Bool ok;
   UChar vbits8;
   ok = get_vbits8(addr, &vbits8);
   if (ok && vbits8 && !tag) {
      VG_(printf)("Missing tag for loaded byte at %llx\n", location);
      tl_assert(!vbits8 || tag);
   }
#endif
   return tag;
}

ULong FC_(helper_LOAD_TAG_2) (Addr addr, ULong addrTag, ULong location) {
   check_enough_tags(3);
   return merge2(get_tag(addr, location), get_tag(addr + 1, location),
		 location);
}

ULong FC_(helper_LOAD_TAG_4) (Addr addr, ULong addrTag, ULong location) {
   check_enough_tags(5);
   return merge4(get_tag(addr, location),     get_tag(addr + 1, location),
		 get_tag(addr + 2, location), get_tag(addr + 3, location),
		 location);
}

ULong FC_(helper_LOAD_TAG_8) (Addr addr, ULong addrTag, ULong location) {
   check_enough_tags(13);
   return merge4(merge2(get_tag(addr, location),     
			get_tag(addr + 1, location), location),
		 merge2(get_tag(addr + 2, location),
			get_tag(addr + 3, location), location + 4),
		 merge2(get_tag(addr + 4, location),
			get_tag(addr + 5, location), location + 8),
		 merge2(get_tag(addr + 6, location),
			get_tag(addr + 7, location), location + 12),
		 location + 16);
}

ULong FC_(helper_LOAD_TAG_16)(Addr addr, ULong addrTag, ULong location) {
   check_enough_tags(21);
   return
      merge4
      (merge4(get_tag(addr, location),      get_tag(addr + 1, location),
	      get_tag(addr + 2, location),  get_tag(addr + 3, location),
	      location),
       merge4(get_tag(addr + 4, location),  get_tag(addr + 5, location),
	      get_tag(addr + 6, location),  get_tag(addr + 7, location),
	      location),
       merge4(get_tag(addr + 8, location),  get_tag(addr + 9, location),
	      get_tag(addr + 10, location), get_tag(addr + 11, location),
	      location),
       merge4(get_tag(addr + 12, location), get_tag(addr + 13, location),
	      get_tag(addr + 14, location), get_tag(addr + 15, location),
	      location),
       location);
}

static void store_tag(Addr addr, ULong dataTag, ULong location,
		      Word in_rollback) {
   if (in_rollback && !FC_(is_escaping)(addr)) {
      dataTag = combine2(dataTag, get_tag(addr, location), location);
   }
   set_tag(addr, dataTag, location);
}

void FC_(helper_STORE_TAG_1) (Addr addr, ULong addrTag, ULong dataTag,
			      ULong location, Word in_rollback) {
#ifdef EVEN_MORE_DEBUG
   /* Assuming the V bits get written first */
   Bool ok;
   UChar vbits8;
   ok = get_vbits8(addr, &vbits8);
   if (ok && vbits8 && !dataTag) {
      VG_(printf)("Missing tag for stored byte at %llx\n", location);
      tl_assert(!vbits8 || dataTag);
   }
#endif
   store_tag(addr, dataTag, location, in_rollback);
}

void FC_(helper_STORE_TAG_2) (Addr addr, ULong addrTag, ULong dataTag,
			      ULong location, Word in_rollback) {
   store_tag(addr,     dataTag, location, in_rollback);
   store_tag(addr + 1, dataTag, location, in_rollback);
}

void FC_(helper_STORE_TAG_4) (Addr addr, ULong addrTag, ULong dataTag,
			      ULong location, Word in_rollback) {
   store_tag(addr,     dataTag, location, in_rollback);
   store_tag(addr + 1, dataTag, location, in_rollback);
   store_tag(addr + 2, dataTag, location, in_rollback);
   store_tag(addr + 3, dataTag, location, in_rollback);
}

void FC_(helper_STORE_TAG_8) (Addr addr, ULong addrTag, ULong dataTag,
			      ULong location, Word in_rollback) {
   store_tag(addr,     dataTag, location, in_rollback);
   store_tag(addr + 1, dataTag, location, in_rollback);
   store_tag(addr + 2, dataTag, location, in_rollback);
   store_tag(addr + 3, dataTag, location, in_rollback);
   store_tag(addr + 4, dataTag, location, in_rollback);
   store_tag(addr + 5, dataTag, location, in_rollback);
   store_tag(addr + 6, dataTag, location, in_rollback);
   store_tag(addr + 7, dataTag, location, in_rollback);
}

void FC_(helper_STORE_TAG_16)(Addr addr, ULong addrTag, ULong dataTag,
			      ULong location, Word in_rollback) {
   int i;
   for (i = 0; i < 16; i++)
      store_tag(addr + i, dataTag, location, in_rollback);
}

static ULong enclose_node = 0;

void mark_special_tags(void) {
   if (leak_cascade)
      mark_tag_in_use(leak_cascade);
   if (enclose_node)
      mark_tag_in_use(enclose_node);
}

void trace_enter_enclose() {
   enclose_node = 0;
}

void trace_leak(ULong tag, Int width, ULong location) {
   tl_assert(tag);
   check_enough_tags(1);
   if (!leak_cascade)
      leak_cascade = get_next_tag();
   new_edge(tag, leak_cascade, width, "global leak", location);
}

void trace_leak_enclosed(ULong tag, Int width, ULong location) {
   tl_assert(width > 0);
   if (!tag) {
      VG_(printf)("Leak (%d bits) with no tag\n", width);
      tl_assert(0);
   }

   if (!enclose_node) {
      check_enough_tags(1);
      enclose_node = get_next_tag();
   }

   new_edge(tag, enclose_node, width, "enclosed leak", location + 41);
}

void trace_enclosed_output(Addr addr, ULong location) {
   ULong old_tag = get_tag(addr, location + 1);
   ULong new_tag;
   check_enough_tags(1);
   new_tag = get_next_tag();
   if (!enclose_node) {
      VG_(printf)("Enclosed output at %p with no node\n", (void *)addr);
      tl_assert(0);
   }
   if (old_tag)
      new_edge(old_tag, new_tag, 8, "old value of enclosed output",
	       location + 2);
   new_edge(enclose_node, new_tag, 8, "enclosed output", location + 3);
   set_single_tag(addr, new_tag);
}

void trace_enclosed_output_region(Addr addr, Int size, ULong location) {
   if (!enclose_node) {
      VG_(printf)("Enclosed output at %p with no node\n", (void *)addr);
      tl_assert(0);
   }
   check_enough_tags(size);
   add_flow_region(addr, size, enclose_node, location);
}

#define CALL_DEPTH 4096
static Addr shadow_call_stack[CALL_DEPTH];
static ULong shadow_stack_hashes[CALL_DEPTH];
static int call_level = 0;

void FC_(helper_WATCH_CALL)(Addr addr) {
   /*int i;*/
   Addr eip = VG_(get_IP)(VG_(get_running_tid)());
   /*for (i = 0; i < call_level; i++)
      VG_(printf)(" ");
      VG_(printf)("Saw call to 0x%llx at 0x%llx\n", (ULong)addr, (ULong)eip);*/
   last_callsite = eip;
   shadow_call_stack[call_level] = eip;
   shadow_stack_hashes[call_level] = callstack_hash;
   call_level++;
   callstack_hash = 3 * callstack_hash + eip;
   /*VG_(printf)("Call stack hash is 0x%llx\n", callstack_hash);*/
   tl_assert(call_level < CALL_DEPTH);
}

void FC_(helper_WATCH_RET)(void) {
   /*int i;*/
   Addr old_callsite;
   --call_level;
   if (call_level < 0)
      call_level = 0;
   old_callsite = shadow_call_stack[call_level];
   callstack_hash = shadow_stack_hashes[call_level];
   /*for (i = 0; i < call_level; i++)
      VG_(printf)(" ");
      VG_(printf)("Saw return, popping 0x%llx\n", (ULong)old_callsite);*/
   last_callsite = old_callsite;
   /*VG_(printf)("Call stack hash is 0x%llx\n", callstack_hash);*/
}

void shadow_stack_iterate(void) {
   callstack_hash += 541; /* perturb by a random value */
   /*VG_(printf)("Iterating call stack hash to 0x%llx\n", callstack_hash);*/
}
